# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from sqlalchemy import Column, create_engine, orm, types
from sqlalchemy.ext.declarative import declarative_base


engine = create_engine("sqlite://")
session = orm.scoped_session(orm.sessionmaker(bind=engine))
Base = declarative_base()
Base.query = session.query_property()


class RouterTestModel(Base):
    __tablename__ = "routertest"
    id = Column(types.Integer(), default=3, primary_key=True)
    text = Column(types.String(length=200))


class RouterTestCompositeKeyModel(Base):
    __tablename__ = "routertestcomposite"
    id = Column(types.Integer(), default=1, primary_key=True)
    other_id = Column(types.Integer(), default=3, primary_key=True)
    text = Column(types.String(length=200))


Base.metadata.create_all(engine)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import six
from rest_framework.fields import IntegerField
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.test import APIRequestFactory
from rest_framework.viewsets import GenericViewSet, ModelViewSet
from rest_witchcraft.mixins import ExpandableQuerySerializerMixin
from rest_witchcraft.serializers import ExpandableModelSerializer, ModelSerializer

from sqlalchemy.orm import joinedload

from django.test import SimpleTestCase

from .models import Vehicle, session
from .test_routers import UnAuthMixin


class DummySerializer(ModelSerializer):
    class Meta(object):
        session = session
        model = Vehicle
        fields = "__all__"


class VehicleOwnerStubSerializer(Serializer):
    id = IntegerField(source="_owner_id")


class VehicleSerializer(ExpandableModelSerializer):
    foo = IntegerField()

    class Meta(object):
        session = session
        model = Vehicle
        expandable_fields = {"owner": VehicleOwnerStubSerializer(source="*", read_only=True), "foo": IntegerField()}
        fields = "__all__"


class DummyViewSet(ExpandableQuerySerializerMixin, GenericViewSet):
    serializer_class = DummySerializer

    def list(self, request, *args, **kwargs):
        return Response()


class ExpandableViewSet(UnAuthMixin, ExpandableQuerySerializerMixin, ModelViewSet):
    serializer_class = VehicleSerializer
    queryset = Vehicle.objects

    def list(self, request, *args, **kwargs):
        r = super(ExpandableViewSet, self).list(request, *args, **kwargs)
        r.data = {"query": six.text_type(self.get_queryset()), "results": r.data}
        return r


class TestDummyViewSet(SimpleTestCase):
    def test_no_queryset(self):
        self.assertIsNone(DummyViewSet().get_query_serializer())


class TestQuerySerializerMixin(SimpleTestCase):
    def setUp(self):
        super(TestQuerySerializerMixin, self).setUp()
        self.rf = APIRequestFactory()

    def test_invalid_query(self):
        view = ExpandableViewSet.as_view(actions={"get": "list"})

        self.assertEqual(view(self.rf.get("/")).status_code, 200)
        self.assertEqual(view(self.rf.get("/", {"expand": "owner"})).status_code, 200)
        self.assertEqual(view(self.rf.get("/", {"expand": "foo"})).status_code, 200)
        self.assertEqual(view(self.rf.get("/", {"expand": "haha"})).status_code, 400)

    def test_no_query_serializer(self):
        view = ExpandableViewSet.as_view(actions={"get": "list"}, serializer_class=DummySerializer)

        self.assertEqual(view(self.rf.get("/")).status_code, 200)

    def test_eagerload_sql(self):
        view = ExpandableViewSet.as_view(actions={"get": "list"})

        self.assertNotIn("LEFT OUTER JOIN", view(self.rf.get("/")).data["query"])
        self.assertNotIn("LEFT OUTER JOIN", view(self.rf.get("/", {"expand": "foo"})).data["query"])

        r = view(self.rf.get("/", {"expand": "owner"}))
        self.assertIn("LEFT OUTER JOIN", r.data["query"])
        self.assertEqual(r.data["query"].count("LEFT OUTER JOIN"), 1)

    def test_already_eagerload(self):
        view = ExpandableViewSet.as_view(
            actions={"get": "list"}, queryset=Vehicle.objects.options(joinedload(Vehicle.owner))
        )

        r = view(self.rf.get("/", {"expand": "owner"}))
        self.assertIn("LEFT OUTER JOIN", r.data["query"])
        # even if we add more joinedloads sqlalchemy should normalize them
        # as that exact path is already joined in base queryset
        self.assertEqual(r.data["query"].count("LEFT OUTER JOIN"), 1)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from enum import Enum

from rest_framework.exceptions import ValidationError
from rest_framework.fields import ChoiceField
from rest_witchcraft.fields import EnumField, HyperlinkedIdentityField, ImplicitExpandableListField, UriField

from django.conf.urls import url
from django.test import SimpleTestCase, override_settings

from .models import Owner
from .models_composite import RouterTestCompositeKeyModel


class SomeEnum(Enum):
    test1 = 1
    test2 = 2


class TestEnumField(SimpleTestCase):
    def test_choices(self):

        field = EnumField(enum_class=SomeEnum)

        self.assertDictEqual(field.choices, {"test1": 1, "test2": 2})

        field = EnumField(choices=SomeEnum)

        self.assertDictEqual(field.choices, {"test1": 1, "test2": 2})

    def test_to_internal_value_works(self):

        field = EnumField(enum_class=SomeEnum)

        self.assertEqual(field.to_internal_value(2), SomeEnum.test2)
        self.assertEqual(field.to_internal_value("test2"), SomeEnum.test2)

        field = EnumField(choices=SomeEnum)

        self.assertEqual(field.to_internal_value(2), SomeEnum.test2)
        self.assertEqual(field.to_internal_value("test2"), SomeEnum.test2)

    def test_to_internal_value_validates(self):

        field = EnumField(enum_class=SomeEnum)

        with self.assertRaises(ValidationError):
            field.to_internal_value("test3")

        with self.assertRaises(ValidationError):
            field.to_internal_value(3)

        field = EnumField(choices=SomeEnum)

        with self.assertRaises(ValidationError):
            field.to_internal_value("test3")

        with self.assertRaises(ValidationError):
            field.to_internal_value(3)

    def test_to_representation_works(self):

        field = EnumField(enum_class=SomeEnum)

        self.assertEqual(field.to_representation(SomeEnum.test1), "test1")

    def test_to_representation_handles_none(self):

        field = EnumField(enum_class=SomeEnum)

        value = field.to_representation(None)

        self.assertIsNone(value)


@override_settings(ROOT_URLCONF=[url(r"^example/(?P<id>.+)/$", lambda: None, name="owner")])
class TestHyperlinkedIdentityField(SimpleTestCase):
    def test_url(self):
        field = HyperlinkedIdentityField(view_name="foo", lookup_url_kwarg="id", lookup_field="id")
        owner = Owner(id=1, first_name="Jon", last_name="Snow")

        url = field.get_url(owner, "owner", None, None)
        self.assertEqual(url, "/example/1/")

    def test_url_not_saved(self):
        field = HyperlinkedIdentityField(view_name="foo")

        self.assertIsNone(field.get_url(Owner(first_name="Jon", last_name="Snow"), "owner", None, None))


@override_settings(ROOT_URLCONF=[url(r"^example/(?P<id>.+)/(?P<other_id>.+)/$", lambda: None, name="owner")])
class TestHyperlinkedIdentityFieldComposite(SimpleTestCase):
    def test_url_composite(self):
        field = HyperlinkedIdentityField(view_name="foo")
        instance = RouterTestCompositeKeyModel(id=1, other_id=2)

        url = field.get_url(instance, "owner", None, None)

        self.assertEqual(url, "/example/1/2/")


@override_settings(ROOT_URLCONF=[url(r"^example/(?P<id>.+)/$", lambda: None, name="owner")])
class TestUriField(SimpleTestCase):
    def test_url(self):
        field = UriField(view_name="foo", lookup_url_kwarg="id", lookup_field="id")
        owner = Owner(id=1, first_name="Jon", last_name="Snow")

        url = field.get_url(owner, "owner", None, None)
        self.assertEqual(url, "/example/1/")


class TestImplicitExpandableListField(SimpleTestCase):
    def test_to_internal_value(self):
        f = ImplicitExpandableListField(child=ChoiceField(choices=["foo", "foo__bar", "bar"]))

        self.assertEqual(f.run_validation(["foo"]), ["foo"])
        self.assertEqual(set(f.run_validation(["foo__bar"])), {"foo__bar", "foo"})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import unittest

from rest_witchcraft.utils import _django_to_drf

from django.core.exceptions import ValidationError


class TestUtils(unittest.TestCase):
    def test_django_to_drf(self):
        self.assertEqual(_django_to_drf("hello"), "hello")
        self.assertEqual(_django_to_drf(["hello"]), ["hello"])
        self.assertEqual(_django_to_drf({"hello": "world"}), {"hello": "world"})
        self.assertEqual(_django_to_drf(ValidationError("hello")), ["hello"])
        self.assertEqual(_django_to_drf(ValidationError({"hello": "world"})), {"hello": ["world"]})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import simplejson as json
from rest_witchcraft import routers, serializers, viewsets

from django.conf.urls import include, url
from django.test import SimpleTestCase, override_settings

from .models_composite import RouterTestCompositeKeyModel, RouterTestModel, session


class RouterTestModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = RouterTestModel
        session = session
        fields = "__all__"


class RouterTestCompositeKeyModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = RouterTestCompositeKeyModel
        session = session
        fields = "__all__"


class UnAuthMixin(object):
    def perform_authentication(self, request):
        return None


class RouterTestViewSet(UnAuthMixin, viewsets.ModelViewSet):
    queryset = RouterTestModel.query
    serializer_class = RouterTestModelSerializer
    lookup_field = "id"
    lookup_url_kwarg = "pk"


class RouterTestCompositeViewSet(UnAuthMixin, viewsets.ModelViewSet):
    queryset = RouterTestCompositeKeyModel.query
    serializer_class = RouterTestCompositeKeyModelSerializer


class RouterTestCompositeCustomRegexViewSet(UnAuthMixin, viewsets.ModelViewSet):
    queryset = RouterTestCompositeKeyModel.query
    serializer_class = RouterTestCompositeKeyModelSerializer
    lookup_url_regex = "(?P<id>[0-9]+)/other/(?P<other_id>[0-9]+)"


router = routers.DefaultRouter()
router.register(r"test", RouterTestViewSet)
router.register(r"testcomposite", RouterTestCompositeViewSet)
router.register(r"testcompositeregex", RouterTestCompositeCustomRegexViewSet)

urlpatterns = [url(r"^", include(router.urls))]


@override_settings(ROOT_URLCONF="tests.test_routers")
class TestDummyDummy(SimpleTestCase):
    def test_assert_when_no_model_found(self):
        class DummyViewSet(UnAuthMixin, viewsets.ModelViewSet):
            pass

        dummy_router = routers.DefaultRouter()

        with self.assertRaises(AssertionError):
            dummy_router.register(r"dummy", DummyViewSet)

    def test_get_lookup_regex_without_model(self):
        class DummyViewSet(UnAuthMixin, viewsets.ModelViewSet):
            @classmethod
            def get_model(cls):
                return None

        dummy_router = routers.DefaultRouter()

        lookup_regex = dummy_router.get_lookup_regex(DummyViewSet)
        self.assertEqual(lookup_regex, "(?P<pk>[^/.]+)")


@override_settings(ROOT_URLCONF="tests.test_routers")
class TestModelRoutes(SimpleTestCase):
    maxDiff = None

    def setUp(self):
        session.add_all(
            [RouterTestModel(id=1, text="router test model 1"), RouterTestModel(id=2, text="router test model 2")]
        )

    def tearDown(self):
        session.rollback()

    def test_list(self):
        resp = self.client.get("/test/")

        self.assertEqual(
            resp.data, [{"id": 1, "text": "router test model 1"}, {"id": 2, "text": "router test model 2"}]
        )

    def test_retrieve(self):
        resp = self.client.get("/test/2/")

        self.assertEqual(resp.data, {"id": 2, "text": "router test model 2"})

    def test_create(self):
        data = json.dumps({"text": "router test model 3"})
        resp = self.client.post("/test/", data=data, content_type="application/json")

        self.assertEqual(resp.data, {"id": 3, "text": "router test model 3"})

    def test_update(self):
        data = {"text": "router test update 2"}
        resp = self.client.put("/test/2/", data=json.dumps(data), content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 2, "text": "router test update 2"})

    def test_patch_update(self):
        data = {"text": "router test update 2"}
        resp = self.client.patch("/test/2/", data=json.dumps(data), content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 2, "text": "router test update 2"})

    def test_delete(self):
        resp = self.client.delete("/test/2/", content_type="application/json")
        self.assertEqual(resp.status_code, 204)


@override_settings(ROOT_URLCONF="tests.test_routers")
class TestCompositeRoutes(SimpleTestCase):
    def setUp(self):
        session.add_all(
            [
                RouterTestCompositeKeyModel(id=1, other_id=1, text="router composite model 1"),
                RouterTestCompositeKeyModel(id=1, other_id=2, text="router composite model 2"),
            ]
        )

    def tearDown(self):
        session.rollback()

    def test_list(self):
        resp = self.client.get("/testcomposite/")

        self.assertEqual(
            resp.data,
            [
                {"id": 1, "other_id": 1, "text": "router composite model 1"},
                {"id": 1, "other_id": 2, "text": "router composite model 2"},
            ],
        )

    def test_retrieve(self):
        resp = self.client.get("/testcomposite/1/2/")

        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router composite model 2"})

    def test_create(self):
        data = json.dumps({"text": "composite test model 3"})
        resp = self.client.post("/testcomposite/", data=data, content_type="application/json")

        self.assertEqual(resp.data, {"id": 1, "other_id": 3, "text": "composite test model 3"})

    def test_update(self):
        data = json.dumps({"text": "router test update 2"})
        resp = self.client.put("/testcomposite/1/2/", data=data, content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router test update 2"})

    def test_patch_update(self):
        data = json.dumps({"text": "router test update 2"})
        resp = self.client.patch("/testcomposite/1/2/", data=data, content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router test update 2"})

    def test_delete(self):
        resp = self.client.delete("/testcomposite/1/2/", content_type="application/json")
        self.assertEqual(resp.status_code, 204)


@override_settings(ROOT_URLCONF="tests.test_routers")
class TestCompositeRoutesWithCustomRegex(SimpleTestCase):
    def setUp(self):
        session.add_all(
            [
                RouterTestCompositeKeyModel(id=1, other_id=1, text="router composite model 1"),
                RouterTestCompositeKeyModel(id=1, other_id=2, text="router composite model 2"),
            ]
        )

    def tearDown(self):
        session.rollback()

    def test_list(self):
        resp = self.client.get("/testcompositeregex/")

        self.assertEqual(
            resp.data,
            [
                {"id": 1, "other_id": 1, "text": "router composite model 1"},
                {"id": 1, "other_id": 2, "text": "router composite model 2"},
            ],
        )

    def test_retrieve(self):
        resp = self.client.get("/testcompositeregex/1/other/2/")

        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router composite model 2"})

    def test_create(self):
        data = json.dumps({"text": "composite test model 3"})
        resp = self.client.post("/testcompositeregex/", data=data, content_type="application/json")

        self.assertEqual(resp.data, {"id": 1, "other_id": 3, "text": "composite test model 3"})

    def test_update(self):
        data = json.dumps({"text": "router test update 2"})
        resp = self.client.put("/testcompositeregex/1/other/2/", data=data, content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router test update 2"})

    def test_patch_update(self):
        data = json.dumps({"text": "router test update 2"})
        resp = self.client.patch("/testcompositeregex/1/other/2/", data=data, content_type="application/json")
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, {"id": 1, "other_id": 2, "text": "router test update 2"})

    def test_delete(self):
        resp = self.client.delete("/testcompositeregex/1/other/2/", content_type="application/json")
        self.assertEqual(resp.status_code, 204)
